xquery version "3.1";

(:~
 : This module expands the motifs, which are encoded as processing instructions,
 : to full TEI elements. By doing this, we create a kind of intermediate format
 : for the TEI files that serves as a basis for the HTML creation.
 : 
 : Each motif is converted to a tei:span encompassing everything that is between
 : the beginning of a comment/processing instruction and its end.
 :)

module namespace me="http://ahikar.sub.uni-goettingen.de/ns/motifs-expansion";
declare namespace tei="http://www.tei-c.org/ns/1.0";

declare function me:main($pages as node()+)
as node()+ {
    me:transform-pis($pages)
    => me:sort-motifs-into-lines()
    => me:expand()
};

declare function me:transform-pis($nodes as node()*)
as node()* {
    for $node in $nodes return
        typeswitch ($node)
        
        case text() return
            $node
            
        case comment() return
            ()
            
        case processing-instruction() return
            element {QName("http://www.tei-c.org/ns/1.0", "motif")} {
                if ($node[not(me:get-motif-type($node) = "")]) then
                    (attribute type {"start"},
                    attribute n {me:get-motif-type($node)},
                    if (me:get-mid($node) != "") then
                        attribute mid {me:get-mid($node)}
                    else
                        ())
                else
                    (attribute type {"end"},
                    if (me:get-mid($node) != "") then
                        attribute mid {me:get-mid($node)}
                    else
                        ())
            }
            
        default return
            element {QName("http://www.tei-c.org/ns/1.0", local-name($node))} {
                $node/@*,
                me:transform-pis($node/node())
            }
};

declare function me:sort-motifs-into-lines($nodes as node()*)
as node()* {
    for $node in $nodes return
        typeswitch ( $node )
        case text() return
            $node
            
        case element(tei:ab) return
            let $previous-line := $node/preceding-sibling::tei:ab[1]
            let $next-line := $node/following-sibling::tei:ab[1]
            let $following-motif-ends-outside-line := $node/following-sibling::tei:motif[@type = "end"]
            let $preceding-motif-starts-outside-line := $node/preceding-sibling::tei:motif[@type = "start"]
            
            return
                if ($next-line and $previous-line
                and $following-motif-ends-outside-line[. << $next-line]
                and $preceding-motif-starts-outside-line[. >> $previous-line]) then
                    element {QName("http://www.tei-c.org/ns/1.0", local-name($node))} {
                        $node/@*,
                        $preceding-motif-starts-outside-line[. >> $previous-line],
                        me:sort-motifs-into-lines($node/node()),
                        $following-motif-ends-outside-line[. << $next-line]
                    }
                
                else if ($next-line
                and $following-motif-ends-outside-line[. << $next-line]) then
                    element {QName("http://www.tei-c.org/ns/1.0", local-name($node))} {
                        $node/@*,
                        me:sort-motifs-into-lines($node/node()),
                        $following-motif-ends-outside-line[. << $next-line]
                    }
                    
                else if ($previous-line
                and $preceding-motif-starts-outside-line[. >> $previous-line]) then
                    element {QName("http://www.tei-c.org/ns/1.0", local-name($node))} {
                        $node/@*,
                        $preceding-motif-starts-outside-line[. >> $previous-line],
                        me:sort-motifs-into-lines($node/node())
                    }
                    
                else
                    me:copy-node-for-sorting($node)
            
        case element(tei:motif) return
            if ($node/ancestor::tei:ab) then
                me:copy-node-for-sorting($node)
            else
                ()
        
        default return
            me:copy-node-for-sorting($node)
};

declare function me:copy-node-for-sorting($node as node())
as node() {
    element {QName("http://www.tei-c.org/ns/1.0", local-name($node))} {
        $node/@*,
        me:sort-motifs-into-lines($node/node())
    }
};

declare function me:get-motif-type($motif as processing-instruction())
as xs:string {
    substring-after($motif, "comment=""")
    => substring-before("""")
};

declare function me:get-mid($node as processing-instruction())
as xs:string {
    substring-after($node, "mid=""")
    => substring-before("""")
};

declare function me:expand($nodes as node()*)
as node()* {
    for $node in $nodes return
        typeswitch ($node)
        
        case text() return
            if (me:is-node-in-motif($node)) then
                ()
            else
                $node
        
        case element(tei:motif) return
            if (me:is-first-motif-of-nested-motifs($node)) then
                me:make-nested-motifs($node)
                
            (: the last line of nested motifs encompassing multiple lines :)
            else if(me:is-first-motif-end-of-nested-motifs($node)) then
                let $motif-start := me:get-previous-motif-start($node, $node/@mid)
                return
                    if (me:is-motif-one-liner($motif-start)) then
                        ()
                    else
                        me:make-nested-motifs-end($node)
                
            (: all (nested) motifs except the first one are skipped in order to
            avoid duplication :)
            else if (me:is-last-motif-of-nested-motifs($node)
            or me:is-inner-motif-of-nested-motif($node)
            or me:is-last-motif-end-of-nested-motifs($node)
            or me:is-inner-motif-end-of-nested-motifs($node)) then
                ()
            
            (: motifs starting and ending in the same line :)
            else if ($node/@type = "start"
            and me:is-motif-one-liner($node)) then
                element {QName("http://www.tei-c.org/ns/1.0","span")} {
                    attribute type {"motif"},
                    $node/@n,
                    let $comment-end := me:get-next-motif-end($node)
                    let $nodes-inbetween := $node/following-sibling::node()[. >> $node and . << $comment-end]
                    return
                        $nodes-inbetween
                }
            
            (: the first line of a motifs encompassing multiple lines :)
            else if($node/@type = "start"
            and not(me:is-motif-one-liner($node))) then
                element {QName("http://www.tei-c.org/ns/1.0","span")} {
                    attribute id {generate-id($node) || "-" || me:determine-motif-part($node, ())},
                    me:make-core-attributes($node),
                    me:make-next-attribute(generate-id($node), $node),
                    let $last-node-of-parent := $node/../node()[last()]
                    let $nodes-of-line-in-motif := ($node/following-sibling::node()[. << $last-node-of-parent], $last-node-of-parent)
                    return
                        $nodes-of-line-in-motif
                }
            
            (: the last line of a motifs encompassing multiple lines :)
            else if($node/@type = "end") then
                let $comment-start := me:get-previous-motif-start($node, ())
                return
                    if (not(me:is-motif-one-liner($comment-start))) then
                        element {QName("http://www.tei-c.org/ns/1.0","span")} {
                            attribute id {generate-id($comment-start) || "-" || me:determine-motif-part($node, ())},
                            me:make-core-attributes($comment-start),
                            let $first-node-of-parent := $node/../node()[1]
                            let $nodes-of-line-in-motif := ($first-node-of-parent,  $first-node-of-parent/following-sibling::node()[. << $node])
                            return
                                $nodes-of-line-in-motif
                        }
                    else
                        ()
                
            else
                ()
                
        case element(tei:ab) return
            (: one of the middles lines of a nested motifs encompassing multiple lines :)
            if (me:is-node-in-nested-motif($node)) then
                element {QName("http://www.tei-c.org/ns/1.0", local-name($node))} {
                    $node/@*,
                    me:make-line-in-nested-motif($node)
                }
                
            (: one of the middles lines of a motifs encompassing multiple lines :)
            else if (me:is-node-in-motif($node)) then
                let $previous-comment-start := me:get-previous-motif-start($node, ())
                let $motif-id := generate-id($previous-comment-start)
                return
                    element {QName("http://www.tei-c.org/ns/1.0", local-name($node))} {
                        $node/@*,
                        element {QName("http://www.tei-c.org/ns/1.0", "span")} {
                            attribute id {$motif-id || "-" || me:determine-motif-part($node, ())},
                            me:make-core-attributes($previous-comment-start),
                            me:make-next-attribute($motif-id, $node),
                            $node/node()
                        }
                    }
            else
                me:copy-node($node)
                
        case element(tei:pb) return
            element {QName("http://www.tei-c.org/ns/1.0", local-name($node))} {
                $node/@*
            }
            
        default return
            me:copy-node($node)
};

declare function me:is-first-motif-of-nested-motifs($node as element(tei:motif))
as xs:boolean {
    $node[@type = "start"]
    and $node/following-sibling::node()[1][self::tei:motif and @type = "start"]
    and not($node/preceding-sibling::node()[1][self::tei:motif and @type = "start"])
};

declare function me:make-nested-motifs($node as element(tei:motif))
as element(tei:span) {
    element {QName("http://www.tei-c.org/ns/1.0","span")} {
        me:make-motif-attributes($node, $node, (), if (me:is-motif-one-liner($node)) then false() else true()),
        if (me:is-last-motif-of-nested-motifs($node)) then
            let $last-node-of-parent := $node/../node()[not(self::tei:motif)][last()]
            let $nodes-of-line-in-motif := ($node/following-sibling::node()[. << $last-node-of-parent], $last-node-of-parent)
            return
                $nodes-of-line-in-motif
        else
            me:make-nested-motifs($node/following-sibling::node()[1][self::tei:motif/@type = "start"])
    }
};

declare function me:make-motif-attributes($node as node(),
    $motif-start as element(tei:motif)?,
    $mid as xs:string?,
    $make-next-attribute as xs:boolean)
as attribute()+ {
    let $base-id := generate-id($motif-start)
    return
        (
            attribute id {$base-id || "-" || me:determine-motif-part($node, $mid)},
            me:make-core-attributes($motif-start),
            if ($make-next-attribute) then
                me:make-next-attribute($base-id, $node)
            else
                ()
        )
};

declare function me:is-last-motif-of-nested-motifs($node as element(tei:motif))
as xs:boolean {
    $node[@type = "start"]
    and $node/preceding-sibling::node()[1][self::tei:motif/@type = "start"]
    and not($node/following-sibling::node()[1][self::tei:motif/@type = "start"])
};

declare function me:is-first-motif-end-of-nested-motifs($node as element(tei:motif))
as xs:boolean {
    $node[@type = "end"]
    and not($node/preceding-sibling::node()[1][self::tei:motif/@type = "end"])
    and $node/following-sibling::node()[1][self::tei:motif/@type = "end"]
};

declare function me:make-nested-motifs-end($node as element(tei:motif))
as element(tei:span) {
    let $mid := $node/@mid/string()
    let $corresponding-motif-start := me:get-previous-motif-start($node, $mid)
    return
        element {QName("http://www.tei-c.org/ns/1.0","span")} {
            me:make-motif-attributes($node, $corresponding-motif-start, $mid, false()),
            if (me:is-last-motif-end-of-nested-motifs($node)) then
                let $first-node-of-parent := $node/../node()[1]
                let $nodes-of-line-in-motif := ($first-node-of-parent, $node/preceding-sibling::node()[. >> $first-node-of-parent][not(self::tei:motif)])
                return
                    $nodes-of-line-in-motif
            else
                me:make-nested-motifs-end($node/following-sibling::node()[1][self::tei:motif/@type = "end"])
        }
};

declare function me:is-last-motif-end-of-nested-motifs($node as element(tei:motif))
as xs:boolean {
    $node[@type = "end"]
    and $node/preceding-sibling::node()[1][self::tei:motif/@type = "end"]
    and not($node/following-sibling::node()[1][self::tei:motif/@type = "end"])
};

declare function me:is-inner-motif-of-nested-motif($node as element(tei:motif))
as xs:boolean {
    $node[@type = "start"]
    and $node/preceding-sibling::node()[1][self::tei:motif/@type = "start"]
};

declare function me:is-inner-motif-end-of-nested-motifs($node as element(tei:motif))
as xs:boolean {
    $node[@type = "end"]
    and $node/preceding-sibling::node()[1][self::tei:motif/@type = "end"]
    and $node/following-sibling::node()[1][self::tei:motif/@type = "end"]
};

declare function me:is-node-in-motif($node as node())
as xs:boolean {
    let $previous-motif-start := me:get-previous-motif-start($node, ())
    let $previous-motif-end := me:get-previous-motif-end($node)
    let $next-motif-start := me:get-next-motif-start($node)
    let $next-motif-end :=  me:get-next-motif-end($node)
    let $are-basic-conditions-fulfilled :=
        $previous-motif-start
        and not($node/descendant::tei:motif[@type = "start"])
        and $node[. >> $previous-motif-start and . << $next-motif-end]
    
    return
        if ($next-motif-start and $previous-motif-end) then
            $are-basic-conditions-fulfilled
            and $next-motif-start[. >> $next-motif-end]
            and $previous-motif-end[. << $previous-motif-start]
            
        else if($next-motif-start) then
            $are-basic-conditions-fulfilled
            and $next-motif-start[. >> $next-motif-end]
            
        else
            $are-basic-conditions-fulfilled
};

declare function me:is-motif-one-liner($node as element(tei:motif))
as xs:boolean {
    let $comment-end := me:get-next-motif-end($node)
    return
        $node/ancestor::tei:ab = $comment-end/ancestor::tei:ab
};

declare function me:copy-node($node as node()*) {
    if (me:is-node-in-motif($node)) then
        ()
    else
        element {QName("http://www.tei-c.org/ns/1.0", local-name($node))} {
            $node/@*,
            me:expand($node/node())
        }
};

declare function me:get-previous-motif-start($node as node(),
    $mid as xs:string?)
as element(tei:motif)? {
    if ($mid) then
        $node/preceding::tei:motif[@type = "start" and @mid = $mid]
    else
        $node/preceding::tei:motif[@type = "start" and not(@mid)][1]
};

declare function me:get-previous-motif-end($node as node())
as element(tei:motif)? {
    $node/preceding::tei:motif[@type = "end"][1]
};

declare function me:get-next-motif-end($node as node())
as element(tei:motif)? {
    $node/following::tei:motif[@type = "end"][1]
};

declare function me:get-next-motif-start($node as node())
as element(tei:motif)? {
    $node/following::tei:motif[@type = "start"][1]
};

declare function me:determine-motif-part($node as node(),
    $mid as xs:string?)
as xs:integer {
    if ($node/@type = "start") then
        1
    else
        let $motif-start := me:get-previous-motif-start($node, ())
        let $motif-line := $motif-start/ancestor::tei:ab
        let $current-line := $node/ancestor-or-self::tei:ab
        return
            (: we have to add 2 here because neither $current-line nor $motif-line
            are considered in the computation :)
            count($current-line/preceding-sibling::tei:ab[. >> $motif-line]) + 2
};

declare function me:is-node-in-nested-motif($node as node())
as xs:boolean {
    me:is-node-in-motif($node)
    and me:is-first-motif-of-nested-motifs(me:get-previous-motif-start($node, ()))
};

declare function me:make-line-in-nested-motif($node as element(tei:ab))
as element(tei:span) {
    let $closest-previous-motif := $node/preceding::node()[self::tei:motif[@type = "start"]][1]
    let $nested-motif-starts := me:get-nested-motif-starts($closest-previous-motif, ())
    return
        me:make-nested-motifs-for-line-in-motifs($node, $nested-motif-starts)
};

declare function me:get-nested-motif-starts($node as element(tei:motif),
    $saved-nested-motifs as element(tei:motif)*)
as element(tei:motif)+ {
    let $previous-nested-motif-start := $node/preceding-sibling::node()[1][self::tei:motif[@type = "start"]]
    return
        if ($previous-nested-motif-start) then
            me:get-nested-motif-starts($previous-nested-motif-start, ($node, $saved-nested-motifs))
        else
            ($node, $saved-nested-motifs)
};

declare function me:make-nested-motifs-for-line-in-motifs($node as element(tei:ab),
    $motif-starts as element(tei:motif)+)
as element(tei:span) {
    let $mid := $motif-starts[1]/@mid
    return
        element {QName("http://www.tei-c.org/ns/1.0","span")} {
            me:make-motif-attributes($node, $motif-starts[1], $mid, true()),
            if (count($motif-starts) = 1) then
                $node/node()
            else
                me:make-nested-motifs-for-line-in-motifs($node, $motif-starts[position() gt 1])
        }
};

declare function me:make-next-attribute($motif-id as xs:string,
    $node as node())
as attribute() {
    attribute next {"#" || $motif-id || "-" || (me:determine-motif-part($node, ()) + 1)}
};

declare function me:make-core-attributes($motif-start as element(tei:motif))
as attribute()+ {
    attribute type {"motif"},
    $motif-start/@n
};
